package controller;

import java.util.ArrayList;
import java.util.HashMap;

import main.EcologiaIO;
import model.Carnivore;
import model.Genome;
import model.Herbivore;
import model.Simulator;

/**
 * The World class acts as a communicator between the model and the view packages. It receives
 * the current status of the simulation from model and passes it on to view. Conversely, user
 * input from view is forwarded to model. It also stores all simulation settings.
 * 
 * @author Daniel Vedder
 * @version 29.8.2014
 */
public class World 
{
	private static World world; //The Singleton instance of this class
	
	//Parameter variables
	private int[] size; //The size of the world (x*y)
	private int timelapse; //When running, the simulation will be updated every so many milliseconds.
	private int stopAt; //The simulation will stop once this update is reached.
	private Humidity humidity; //The humidity level (from -2 to 2)
	private int startGrassDensity; //The initial grass density on all fields
	private int startNoWaterTiles; //The number of water tiles that will be created.
	private int startNoCarnivores, startNoHerbivores; //The starting number of carnivores/herbivores.
	private int startEnergyCarnivores, startEnergyHerbivores; //The starting energy for carnivores/herbivores.
	private int killRate; //A percentage probability for how often carnivore attacks will be successful.

	//Runtime variables
	private boolean running; //Is the simulation running?
	private int turn; //The update number
	private int nextID; //The next ID number that will be handed out to a newborn animal
	private int herbivoreCounter, carnivoreCounter; //Keep count of the herbivores and carnivores
	private int highestGeneration; //What generation have we reached by now?
	private int averageGrassDensity; //A measure of how much food is available for the herbivores
	private ArrayList<HashMap<String, Integer>> animals; //A list of properties of each animal
	private ArrayList<String> news; //A collection of news items that have accumulated
	
	/**
	 * This class implements Singleton, therefore the constructor is private.
	 * TODO: load in settings from a config file
	 */
	private World()
	{
		//Parameter settings
		size = new int[] {100, 100}; //Default: 100*100
		timelapse = 100; //Default: 300 - can be changed at run-time
		stopAt = 200; //Default: 100 - can be changed at run-time
		startNoWaterTiles = 10; //Default: 10
		humidity = Humidity.WET; //Default: Humidity.WET - can be changed at run-time
		startGrassDensity = 100; //Default: 100
		startNoCarnivores = 50; //Default: 15 - Hypothetical ideal: 5
		startNoHerbivores = 200; //Default: 30 - Hypothetical ideal: 160
		startEnergyCarnivores = 150; //Default: 150
		startEnergyHerbivores = 100; //Default: 100
		killRate = 30; //Default: 30

		//readConfigFile();
		
		//Runtime variables
		running = false;
		turn = 0;
		nextID = 1;
		herbivoreCounter = 0;
		carnivoreCounter = 0;
		highestGeneration = 1;
		averageGrassDensity = startGrassDensity;
		animals = null;
		news = new ArrayList<String>();
	}
	
	/**
	 * The Singleton method.
	 */
	public static World getInstance()
	{
		if (world == null) {
			world = new World();
		}
		return world;
	}
	
	/**
	 * Reset the world run-time variables, ready for a new run.
	 * This method should only be called from the Ecologia main class!
	 */
	public void reset()
	{
		running = false;
		turn = 0;
		herbivoreCounter = 0;
		carnivoreCounter = 0;
		animals = null;
		news = new ArrayList<String>();
	}
	
	/**
	 * Display a news item - calling with null as a parameter resets the news list
	 * @param news
	 */
	public void giveNews(String message)
	{
		if (message == null) {
			news.clear();
		}
		else {
			message = turn+": "+message;
			news.add(message);
			EcologiaIO.log(message);
		}
	}
	
	/**
	 * Return information about the animal at the given position as a hash map
	 * @param x, y
	 * @return HashMap, or null if no animal at the specified location
	 */
	public HashMap<String, Integer>	getAnimalInfo(int x, int y)
	{
		HashMap<String, Integer> info = null;
		for (int a = 0; a < animals.size(); a++) {
			if (animals.get(a).get("X") == x && animals.get(a).get("Y") == y) {
				info = animals.get(a);
				break;
			}
		}
		return info;
	}
	
	/**
	 * Return information about the map field at the given position as a hash map
	 * @param x, y
	 * @return HashMap, or null if out of bounds
	 */
	public HashMap<String, Integer>	getFieldInfo(int x, int y)
	{
		return Simulator.getField(x, y).getInfo();
	}

	/*
	 * All the getters and setters for the parameter settings and runtime variables
	 */

	/**
	 * Return a hash map holding all the genome values
	 */
	public HashMap<String, Integer> getDefaultGenome(OccupantType type)
	{
		HashMap<String, Integer> genomeInfo = new HashMap<String, Integer>();
		Genome genome = null;
		if (type == OccupantType.HERBIVORE) genome = Herbivore.defaultGenome;
		else if (type == OccupantType.CARNIVORE) genome = Carnivore.defaultGenome;
		genomeInfo.put("Mutation rate", genome.getMutationRate());
		genomeInfo.put("Speed", genome.getSpeed());
		genomeInfo.put("Stamina", genome.getStamina());
		genomeInfo.put("Sight", genome.getSight());
		genomeInfo.put("Metabolism", genome.getMetabolism());
		genomeInfo.put("Age limit", genome.getAgeLimit());
		genomeInfo.put("Strength", genome.getStrength());
		genomeInfo.put("Reproductive energy", genome.getReproductiveEnergy());
		genomeInfo.put("Maturity age", genome.getMaturityAge());
		genomeInfo.put("Gestation", genome.getGestation());
		genomeInfo.put("Reproduction rate", genome.getReproductionRate());
		return genomeInfo;
	}
	
	/**
	 * Interface for the Genome method
	 */
	public void setDefaultGenome(OccupantType type, int mutationRate, int speed, int stamina,
								 int sight, int metabolism, int ageLimit, int strength,
								 int reproductiveEnergy, int maturityAge, int gestation,
								 int reproductionRate)
	{
		Genome genome = new Genome(mutationRate, speed, stamina, sight, metabolism,
								   ageLimit, strength, reproductiveEnergy, maturityAge,
								   gestation, reproductionRate);
		if (type == OccupantType.HERBIVORE) Herbivore.defaultGenome = genome;
		else if (type == OccupantType.CARNIVORE) Carnivore.defaultGenome = genome;
	}
	
	public int[] getSize() 
	{
		return size;
	}

	public void setSize(int[] size) 
	{
		this.size = size;
	}

	public int getTimelapse() 
	{
		return timelapse;
	}

	public void setTimelapse(int timelapse) 
	{
		if (this.timelapse != timelapse) {
			EcologiaIO.log("Timelapse changed to "+timelapse+" ms.");
		}
		this.timelapse = timelapse;
	}

	public int getStopAt() 
	{
		return stopAt;
	}

	public void setStopAt(int stopAt) 
	{
		this.stopAt = stopAt;
	}
	
	public Humidity getHumidity() 
	{
		return humidity;
	}

	public void setHumidity(Humidity humidity) 
	{
		this.humidity = humidity;
	}

	public int getStartGrassDensity()
	{
		return startGrassDensity;
	}

	public void setStartGrassDensity(int startGrassDensity)
	{
		this.startGrassDensity = startGrassDensity;
	}

	public int getStartNoWaterTiles()
	{
		return startNoWaterTiles;
	}

	public void setStartNoWaterTiles(int startNoWaterTiles)
	{
		this.startNoWaterTiles = startNoWaterTiles;
	}

	public int getStartNoCarnivores()
	{
		return startNoCarnivores;
	}

	public void setStartNoCarnivores(int startNoCarnivores)
	{
		this.startNoCarnivores = startNoCarnivores;
	}

	public int getStartNoHerbivores()
	{
		return startNoHerbivores;
	}

	public void setStartNoHerbivores(int startNoHerbivores)
	{
		this.startNoHerbivores = startNoHerbivores;
	}

	public int getStartEnergyCarnivores() 
	{
		return startEnergyCarnivores;
	}

	public void setStartEnergyCarnivores(int startEnergyCarnivores) 
	{
		this.startEnergyCarnivores = startEnergyCarnivores;
	}

	public int getStartEnergyHerbivores() 
	{
		return startEnergyHerbivores;
	}

	public void setStartEnergyHerbivores(int startEnergyHerbivores) 
	{
		this.startEnergyHerbivores = startEnergyHerbivores;
	}

	public int getKillRate()
	{
		return killRate;
	}

	public void setKillRate(int killRate)
	{
		this.killRate = killRate;
	}

	public int getHerbivoreCount() 
	{
		return herbivoreCounter;
	}

	public void setHerbivoreCount(int herbivoreCounter) 
	{
		this.herbivoreCounter = herbivoreCounter;
	}

	public int getCarnivoreCount() 
	{
		return carnivoreCounter;
	}

	public void setCarnivoreCount(int carnivoreCounter) 
	{
		this.carnivoreCounter = carnivoreCounter;
	}

	public int getAverageGrassDensity() 
	{
		return averageGrassDensity;
	}

	public void setAverageGrassDensity(int averageGrassDensity) 
	{
		this.averageGrassDensity = averageGrassDensity;
	}

	public boolean isRunning() 
	{
		return running;
	}

	public void setRunning(boolean running) 
	{
		this.running = running;
	}

	public int getTurn() 
	{
		return turn;
	}
	
	/**
	 * Increment the turn variable by one.
	 */
	public void incrementTurn()
	{
		turn++;
	}
	
	public int getNextID()
	{
		int id = nextID;
		nextID++;
		return id;
	}

	public void incGeneration(int n)
	{
		if (n > highestGeneration) {
			highestGeneration = n;
		}
	}

	public int getGeneration()
	{
		return highestGeneration;
	}
	
	public void setAnimals(ArrayList<HashMap<String, Integer>> animalInfo)
	{
		animals = animalInfo;
	}
	
	public ArrayList<String> collectNews()
	{
		return news;
	}
}
