package main;

import controller.World;
import view.GUI;
import model.Simulator;

/**
 * Ecologia is a relatively simple ecosystem simulator, designed to show basic relationships between predators, 
 * prey and producers.
 * 
 * This is the main class, which launches the application and acts as a manager, updating the program each round.
 * 
 * @author Daniel Vedder
 * @version 28.8.2014
 */
public class Ecologia implements Runnable 
{
	public static Ecologia eco; //The singleton object
	public static final String version = "1.0";
	
	private GUI gui;
	private Simulator simulator;
	private EcoTest tester;
	
	private Thread runningThread;
	
	/**
	 * Launch the programme.
	 * see printHelp() for possible arguments
	 */
	public static void main(String[] args) 
	{
		//Print the version number and exit
		if (argumentPassed("--version", args) || argumentPassed("-V", args)) {
			System.out.println("Ecologia "+version);
			System.exit(0);
		}
		
		//Print a short help text and exit
		if (argumentPassed("--help", args) || argumentPassed("-h", args)) {
			printHelp();
			System.exit(0);
		}
		
		//Set runtime flags
		if (argumentPassed("--debug", args) || argumentPassed("-d", args)) EcologiaIO.debugging = true;
		if (argumentPassed("--logging", args) || argumentPassed("-l", args)) EcologiaIO.logging = true;
		if (argumentPassed("--verbose", args) || argumentPassed("-v", args)) EcologiaIO.verbose = true;
		
		if (EcologiaIO.logging) EcologiaIO.archiveLog();
		EcologiaIO.printStatus();
		
		//Create an instance
		eco = new Ecologia();
	}
	
	/**
	 * The Singleton method.
	 */
	public static Ecologia getInstance()
	{
		return eco;
	}
	
	/**
	 * Ecologia implements Singleton, so the constructor is private.
	 */
	private Ecologia()
	{
		EcologiaIO.log("Launching Ecologia...");
		simulator = new Simulator();
		gui = new GUI();
		tester = new EcoTest();
		EcologiaIO.debug("Launch completed.");
	}
	
	/**
	 * Reset the simulator in order to start a new run.
	 */
	public void reset()
	{
		EcologiaIO.archiveLog();
		EcologiaIO.log("Resetting Ecologia...");
		World.getInstance().reset();
		simulator = null;
		simulator = new Simulator();
		gui.dispose();
		gui = null;
		gui = new GUI();
		gui.update();
	}
	
	/**
	 * Start the simulation.
	 */
	public void startThread()
	{
  	  World.getInstance().setRunning(true);
  	  runningThread = new Thread(this);
      runningThread.start();
	}

	/**
	 * Run the simulation.
	 */
	public void run() 
	{
		World.getInstance().giveNews("Simulation is running.");
		while (World.getInstance().isRunning()) {
			iterate();
		}
		World.getInstance().giveNews("Simulation has stopped.");
		gui.update(); //Make sure the above news is displayed by the GUI
	}
	
	/**
	 * Perform one iteration of the simulation.
	 */
	public synchronized void iterate()
	{
		World.getInstance().incrementTurn();
		EcologiaIO.log("Executing update "+World.getInstance().getTurn());
		if (EcologiaIO.debugging) {tester.runTest();}
		simulator.update();
		//If the stopAt number is reached, pause the simulation
		if (World.getInstance().getStopAt() == World.getInstance().getTurn()) {
			World.getInstance().setRunning(false);
		}
		gui.update();
		EcologiaIO.log("Herbivore count: "+World.getInstance().getHerbivoreCount());
		EcologiaIO.log("Carnivore count: "+World.getInstance().getCarnivoreCount());		
		//Stop the simulation if there are no more animals
		if (World.getInstance().getCarnivoreCount() == 0 && World.getInstance().getHerbivoreCount() == 0) {
			World.getInstance().setRunning(false);
			return;
		}
		//Pause for as long as the user wants (default = 300 milliseconds)
		try {
			Thread.sleep(World.getInstance().getTimelapse());
		}
		catch (InterruptedException ie) {}
	}
	
	/**
	 * Print a short help text when invoked from the commandline
	 */
	private static void printHelp()
	{
		System.out.println("Ecologia "+version+", the simple ecosystem simulator.");
		System.out.println("Licensed under the terms of the GNU General Public License v3");
		System.out.println("\nCommandline options:");
		System.out.println("--help    -h	Print this help text");
		System.out.println("--version -V	Print the version number");
		System.out.println("--verbose -v	Give verbose output");
		System.out.println("--logging -l	Enable logging");
		System.out.println("--debug   -d	Set the debug flag\n");
	}
	
	/**
	 * Is the argument str to be found in the array args?
	 * @param str
	 * @param args
	 */
	private static boolean argumentPassed(String str, String[] args)
	{
		for (int a = 0; a < args.length; a++) {
			if (args[a].equals(str)) {
				return true;
			}
		}
		return false;
	}

}
